#include <fstream>
#include <FlexLexer.h>
#include "kite.h"
#include "KiteVM.h"
#include "KiteObject.h"
#include "KiteCode.h"
#include "KiteIdentifier.h"
#include "KiteBoolean.h"
#include "KiteString.h"
#include "KiteList.h"
#include "KiteKeyValPair.h"
#include "KiteTree.h"
#include "KiteFile.h"
#include "KiteDynamicObject.h"
#include "KiteMath.h"
#include "KiteSocket.h"
#include "KiteSystem.h"
#include "KiteExternalFunction.h"

namespace KiteVM {
	KiteReference<KiteObject> LoadInstruction::eval(VM &vm) {
		KiteReference<KiteObject> name(vm.stack.top());
		vm.stack.pop();
        
        fstream file;
        file.open(((KiteString&)(*name)).getString().c_str(), ios::in);
        
        yyFlexLexer lexer(&file);
        ParserState state(&lexer);
        void *pParser;
        int hTokenId;
        state.file = ((KiteString&)(*name)).getString();

        try {        
        pParser = ParseAlloc( malloc );
        while( (hTokenId = lexer.yylex()) != -1 ){
            if (hTokenId == NUMBER || hTokenId == STRING || hTokenId == IDENTIFIER) state.text = lexer.YYText();
            Parse(pParser, hTokenId, NULL, &state);
        }
        Parse(pParser, 0, NULL, &state);
        ParseFree(pParser, free );
     
        KiteVM::Instruction *i = state.root;          
        if (i) {
            vm.run(i);
            i->deallocate();
            return vm.stack.top();
        } else {
            return KiteReference<KiteObject>(new KiteBoolean(false));
        }
        } catch (KiteReference<KiteObject> &exc) {
            // Syntax error
            return exc;
        }
    }

	KiteReference<KiteObject> VarInstruction::eval(VM &vm) {
		KiteReference<KiteIdentifier> lhs(vm.stack.top().referenceIdentifier());
		vm.stack.pop();
		KiteReference<KiteObject> rhs(vm.stack.top());
		vm.stack.pop();

		vm.syms.addVar(((KiteIdentifier&)(*lhs)).getString(), rhs);
		return rhs;
	}

	KiteReference<KiteObject> ClassInstruction::eval(VM &vm) {
		KiteReference<KiteObject> name(vm.stack.top());
		vm.stack.pop();
		KiteReference<KiteObject> definition(vm.stack.top());
		vm.stack.pop();

		if (typeid(*definition) == typeid(KiteDynamicObject)) {
			// Simple assignment
			vm.syms.addVar((*name.referenceIdentifier()).getString(), definition);
            ((KiteDynamicObject&)(*definition)).type = (*name.referenceIdentifier()).getString();
            return definition;
		} else if (typeid(*definition) == typeid(KiteCode)) {
			KiteCode &c = (KiteCode&)(*definition);

			// Create class
			KiteReference<KiteObject> ret(new KiteDynamicObject(vm, c.getCode()));
			if (typeid(*ret) != typeid(KiteTree)) {
				vm.syms.addVar((*name.referenceIdentifier()).getString(), ret);
			}
            
            ((KiteDynamicObject&)(*ret)).type = (*name.referenceIdentifier()).getString();
			return ret;
		}
        
        return KiteReference<KiteObject>(new KiteTree("Class requires a code object"));
	}

    const int MAX_FUNCTION_CALLS = 5000;
    
	KiteReference<KiteObject> FunctionCallInstruction::eval(VM &vm) {
		KiteReference<KiteObject> lhs(vm.stack.top());
		vm.stack.pop();
		KiteReference<KiteObject> name(vm.stack.top());
		vm.stack.pop();
		KiteReference<KiteObject> args(vm.stack.top());
		vm.stack.pop();

		KiteReference<KiteObject> newlhs, newrhs;
		if (typeid(*lhs) == typeid(KiteIdentifier) &&
            ((KiteIdentifier&)(*lhs)).getString() != "tree" &&
            ((KiteIdentifier&)(*lhs)).getString() != "file") {
			// Gotta find the lhs in the symbol table first
			newlhs = vm.syms.get((*lhs.referenceIdentifier()).getString());
		} else {
			newlhs = lhs;
		}

		if (typeid(*args) == typeid(KiteIdentifier)) {
			// Gotta find the rhs in the symbol table first
			newrhs = vm.syms.get((*args.referenceIdentifier()).getString());
		} else {
			newrhs = args;
		}

		// If a class object (and new was called), create new instance
		if ((*name.referenceIdentifier()).getString() == "new") {
            if (typeid(*lhs) == typeid(KiteIdentifier) &&
                ((KiteIdentifier&)(*lhs)).getString()  == "tree") {
                string msg = ((KiteString&)(*coerceString(vm, newrhs))).getString();
                vm.funcCalls.push_back(FuncCall("tree|new", this->file, this->line));
                KiteReference<KiteObject> ret(new KiteTree(msg));
                ((KiteTree&)(*ret)).setBacktrace(vm.funcCalls);
                vm.funcCalls.pop_back();
                return ret;
            } else if (typeid(*lhs) == typeid(KiteIdentifier) &&
                ((KiteIdentifier&)(*lhs)).getString() == "file") {
                vm.funcCalls.push_back(FuncCall("file|new", this->file, this->line));
                KiteReference<KiteObject> ret(new KiteFile(vm, args));
                if (typeid(*ret) == typeid(KiteTree)) {
                    ((KiteTree&)(*ret)).setBacktrace(vm.funcCalls);
                }
                vm.funcCalls.pop_back();
                return ret;
            }
			KiteReference<KiteObject> ret((*newlhs).create(vm, newlhs, args));
			return ret;
        } else if ((*name.referenceIdentifier()).getString() == "op+"
                 && !(*newlhs).isInstance()
                 && typeid(*newrhs) == typeid(KiteCode)) {
            // New class with LHS as parent class
            KiteReference<KiteObject> ret(new KiteDynamicObject(vm, newlhs, ((KiteCode&)(*newrhs)).getCode()));
            return ret;
		} else {
			vm.syms.addContext();
            vm.syms.addVar("this", newlhs);
            
            // Add to function stack
            if (typeid(*lhs) == typeid(KiteIdentifier)) {
                vm.funcCalls.push_back(FuncCall(((KiteIdentifier&)(*lhs)).getString() + "|" + ((KiteIdentifier&)(*name)).getString(), this->file, this->line));
            } else {
                vm.funcCalls.push_back(FuncCall(((KiteIdentifier&)(*name)).getString(), this->file, this->line));
            }
            bool ignore = false;
            KiteReference<KiteObject> ret;
            KiteReference<KiteObject> fun;
            if (newlhs.getObj() == NULL) {
                ret = new KiteTree("Cannot find identifier: " + ((KiteIdentifier&)(*lhs)).getString());
                goto error_end;
            }
            
            if (   typeid(*newlhs) == typeid(KiteList)
                /*|| typeid(*newlhs) == typeid(KiteString)
                || typeid(*newlhs) == typeid(KiteNumber)
                || typeid(*newlhs) == typeid(KiteBoolean)*/) {
                ignore = true;
                vm.funcCalls.pop_back();
            }
            
            // Dont' do the function call if stack will overflow
            fun = ((*newlhs).dereference(vm, (*name.referenceIdentifier()).getString()));
            
            if (fun.getObj() == NULL) {
                // Cannot find identifier, throw a tree
                if (typeid(*lhs) == typeid(KiteIdentifier)) {
                    ret = new KiteTree("Cannot find identifier or method: " + ((KiteIdentifier&)(*lhs)).getString());
                } else {
                    ret = new KiteTree("Cannot find identifier or method");
                }
            } else if (vm.funcCalls.size() > MAX_FUNCTION_CALLS) {
                ret = new KiteTree("Stack overflow");
            } else {
                ret = (*fun).call(vm, newrhs);
            }
            
error_end:
            if (typeid(*ret) == typeid(KiteTree)) {
                KiteTree &exc = (KiteTree&)(*ret);
                if (exc.getBacktrace().size() == 0) {
                    exc.setBacktrace(vm.funcCalls);
                }
            }
            
            if (!ignore && vm.funcCalls.size() > 0) vm.funcCalls.pop_back();
			vm.syms.removeContext();
			return ret;
		}
	}

	KiteReference<KiteObject> StringInstruction::eval(VM &vm) {
        static map<std::string, KiteReference<KiteObject> > cache;
        if (cache.count(constant) > 0) return cache[constant];
        return KiteReference<KiteObject>(new KiteString(constant));
	}

	KiteReference<KiteObject> IdentifierInstruction::eval(VM &vm) {
        static map<std::string, KiteReference<KiteObject> > cache;
        if (cache.count(constant) > 0) return cache[constant];
        return KiteReference<KiteObject>(new KiteIdentifier(constant));
	}

	KiteReference<KiteObject> BooleanInstruction::eval(VM &vm) {
        static map<std::string, KiteReference<KiteObject> > cache;
        if (cache.count(constant) > 0) return cache[constant];
        return KiteReference<KiteObject>(new KiteBoolean(constant));
	}

	KiteReference<KiteObject> CodeInstruction::eval(VM &vm) {
        static map<Instruction *, KiteReference<KiteObject> > cache;
        if (cache.count(constant) > 0) return cache[constant];
        return (cache[constant] = KiteReference<KiteObject>(new KiteCode(constant)));
	}

	KiteReference<KiteObject> NumberInstruction::eval(VM &vm) {
        static map<std::string, KiteReference<KiteObject> > cache;
        if (cache.count(constant) > 0) return cache[constant];
        return (cache[constant] = KiteReference<KiteObject>(new KiteNumber(constant)));
	}

	KiteReference<KiteObject> AssociateArrayInstruction::eval(VM &vm) {
		KiteReference<KiteObject> key(vm.stack.top());
		vm.stack.pop();
		KiteReference<KiteObject> val(vm.stack.top());
		vm.stack.pop();

		return KiteReference<KiteObject>(new KiteKeyValPair(key, val));
	}

	KiteReference<KiteObject> AssignInstruction::eval(VM &vm) {
		KiteReference<KiteObject> lhs(vm.stack.top());
		vm.stack.pop();
		KiteReference<KiteObject> rhs(vm.stack.top());
		vm.stack.pop();

		if (typeid(*lhs) == typeid(KiteIdentifier)) {
			return (vm.syms.replace((*lhs.referenceIdentifier()).getString(), rhs));
		} else {
			lhs.replace(rhs);
			return lhs;
		}
	}

	KiteReference<KiteObject> ArrayDereferenceInstruction::eval(VM &vm) {
		KiteReference<KiteObject> name(vm.stack.top());
		vm.stack.pop();
		KiteReference<KiteObject> idx(vm.stack.top());
		vm.stack.pop();

		KiteReference<KiteObject> newlhs, newrhs;
		if (typeid(*name) == typeid(KiteIdentifier)) {
			// Gotta find the lhs in the symbol table first
			newlhs = vm.syms.get((*name.referenceIdentifier()).getString());
		} else {
			newlhs = name;
		}

		if (typeid(*idx) == typeid(KiteIdentifier)) {
			// Gotta find the lhs in the symbol table first
			newrhs = vm.syms.get((*idx.referenceIdentifier()).getString());
		} else {
			newrhs = idx;
		}

		vm.syms.addContext();
		vm.syms.addVar("this", newlhs);
		KiteReference<KiteObject> fun((*newlhs).dereference(vm, "opderef"));
		KiteReference<KiteObject> ret((*fun).call(vm, newrhs));
		vm.syms.removeContext();
		return ret;
	}

	KiteReference<KiteObject> EmptyArrayInstruction::eval(VM &vm) {
		return KiteReference<KiteObject>(new KiteList(false));
	}

	KiteReference<KiteObject> MergeArrayInstruction::eval(VM &vm) {
		KiteReference<KiteObject> lhs(vm.stack.top());
		vm.stack.pop();
		KiteReference<KiteObject> rhs(vm.stack.top());
		vm.stack.pop();

		if (typeid(*lhs) == typeid(KiteList) && typeid(*rhs) != typeid(KiteList)) {
			KiteList list1((KiteList&)(*lhs));
			KiteList list2(rhs);
			return KiteReference<KiteObject>(list1.merge(list2));
		} else if (typeid(*rhs) == typeid(KiteList) && typeid(*lhs) != typeid(KiteList)) {
			KiteList list2(((KiteList&)(*rhs)));
			KiteList list1(lhs);
			return KiteReference<KiteObject>(list1.merge(list2));
		} else if (typeid(*rhs) != typeid(KiteList) && typeid(*lhs) != typeid(KiteList)) {
			KiteList list2(rhs);
			KiteList list1(lhs);
			return KiteReference<KiteObject>(list1.merge(list2));
		} else {
			KiteList list1(((KiteList&)(*lhs)));
			KiteList list2(((KiteList&)(*rhs)));
			return KiteReference<KiteObject>(list1.merge(list2));
		}
	}

	KiteReference<KiteObject> DereferenceInstruction::eval(VM &vm) {
		KiteReference<KiteObject> lhs(vm.stack.top());
		vm.stack.pop();
		KiteReference<KiteObject> rhs(vm.stack.top());
		vm.stack.pop();

		if (typeid(*lhs) == typeid(KiteIdentifier)) {
			// Gotta find the lhs in the symbol table first
			KiteReference<KiteObject> newlhs(vm.syms.get((*lhs.referenceIdentifier()).getString()));
			return (*newlhs).dereference(vm, (*rhs.referenceIdentifier()).getString());
		}

		return (*lhs).dereference(vm, (*rhs.referenceIdentifier()).getString());
	}

	KiteReference<KiteObject> VM::run(KiteVM::Instruction *instr) {
		while(instr) {
            //cout << instr->toString() << endl;
			stack.push(instr->eval(*this));

			if (typeid(*stack.top()) == typeid(KiteTree)) {
				break;
			}

			instr = instr->next;
		}

		return stack.top();
	}

	VM::~VM() {
		// nothing required
	}

	VM::VM() {
        //KiteObject::getClassinfo();
        syms.addVar("math", KiteReference<KiteObject>(new KiteMath()));
        syms.addVar("socket", KiteReference<KiteObject>(new KiteSocket()));
        syms.addVar("system", KiteReference<KiteObject>(new KiteSystem()));
        syms.addVar("loader", KiteReference<KiteObject>(new KiteExternalFunction()));
	}
};

KiteReference<KiteObject> coerceNumber(KiteVM::VM &vm, KiteReference<KiteObject> obj) {
	vm.syms.addContext();
	vm.syms.addVar("this", obj);
	KiteReference<KiteObject> fun((*obj).dereference(vm, "number"));
	KiteReference<KiteObject> ret((*fun).call(vm, KiteReference<KiteObject>(NULL)));
	vm.syms.removeContext();
	return ret;
}

KiteReference<KiteObject> coerceString(KiteVM::VM &vm, KiteReference<KiteObject> obj) {
	vm.syms.addContext();
	vm.syms.addVar("this", obj);
	KiteReference<KiteObject> fun((*obj).dereference(vm, "string"));
	KiteReference<KiteObject> ret((*fun).call(vm, KiteReference<KiteObject>(NULL)));
	vm.syms.removeContext();
	return ret;
}

bool valuesEqual(KiteVM::VM &vm, KiteReference<KiteObject> x, KiteReference<KiteObject> j) {
	vm.syms.addContext();
	vm.syms.addVar("this", x);
	KiteReference<KiteObject> fun((*x).dereference(vm, "op=="));
	KiteReference<KiteObject> ret((*fun).call(vm, j));
	vm.syms.removeContext();

	if (typeid(*ret) == typeid(KiteTree)) {
		// Problem encountered; not equal.
		return false;
	} else if (typeid(*ret) == typeid(KiteBoolean)) {
		return ((KiteBoolean&)(*ret)).boolVal();
	} else {
		// Unknown error
		return false;
	}
}
