#include "KiteVM.h"
#include "KiteObject.h"
#include "KiteCode.h"
#include "KiteString.h"
#include "KiteTree.h"
#include "KiteIdentifier.h"
#include "KiteList.h"
#include "KiteDynamicObject.h"
#include "KiteDynamicLoader.h"

using namespace KiteVM;

KiteDynamicObject::KiteDynamicObject(KiteVM::VM &vm, KiteReference<KiteObject> parent, KiteVM::Instruction *ip)
: KiteObject(true), inherits(parent) {
	instance = false;
    classInfoInit(vm, ip);
}

void KiteDynamicObject::classInfoInit(KiteVM::VM &vm, KiteVM::Instruction *ip) {
    // Execute code to create static values
	vm.syms.addContext();
	KiteReference<KiteObject> test(vm.run(ip));
	if (typeid(*test) == typeid(KiteTree)) {
		// Error while running construction code
		throw test;
	}

	// Save symbols created
	syms = vm.syms.getTop();
	vm.syms.removeContext();
}

KiteDynamicObject::KiteDynamicObject(KiteVM::VM &vm, KiteVM::Instruction *ip) : KiteObject(true), inherits(KiteObject::getClassinfo()) {
	instance = false;
	classInfoInit(vm, ip);
}

KiteDynamicObject::KiteDynamicObject(KiteVM::VM &vm, KiteReference<KiteObject> base, KiteReference<KiteObject> param) 
: KiteObject(true), classinfo(base) {
	instance = true;
	type = (*base).type;
}

KiteDynamicObject::KiteDynamicObject(KiteVM::VM &vm, const char *file)
: KiteObject(true), inherits(KiteObject::getClassinfo()), loader(new KiteDynamicLoader(file)) {
	instance = false;

	// Pull filename only, w/o extension, and use it as the type name
	string name(file);
	string newname = name.substr(name.rfind("/"));
	type = newname.substr(0, newname.find("."));

	// Call initializer function
	try {
		KiteReference<KiteObject> (*func)(VM &, KiteReference<KiteObject>) = 
		    (KiteReference<KiteObject> (*)(VM &, KiteReference<KiteObject>))loader->findFunction((type + string("_") + "init").c_str());
		(*func)(vm, NULL);
	} catch (KiteReference<KiteObject> &exc) {
		// Exception caught, ignore
	}
}

KiteReference<KiteObject>
KiteDynamicObject::create(KiteVM::VM &vm, KiteReference<KiteObject> base, KiteReference<KiteObject> param) {
	KiteReference<KiteObject> ret(new KiteDynamicObject(vm, base, param));

	// Call constructor
	vm.syms.addContext();
	vm.syms.addVar("this", ret);
	KiteReference<KiteObject> fun = (*ret).dereference(vm, "new");
	(void)(*fun).call(vm, param); // Call constructor
	vm.syms.removeContext();

	return ret;
}

KiteDynamicObject::~KiteDynamicObject() {
	// Everything will be deconstructed by default, except for the DynamicLoader
	if (loader) {
		// Call destroyer function
		try {
			void (*func)() = (void (*)())loader->findFunction((type + string("_") + "destroy").c_str());
			(*func)();
		} catch (KiteReference<KiteObject> &exc) {
			// Exception caught, ignore
		}
		delete loader;
	}
}

KiteReference<KiteObject> KiteDynamicObject::dereference(VM &vm, string id) {
	// Go through class hierachy, looking for object
	KiteObject *cur = this;
	bool staticobjs = (isInstance() ? false : true);

	while(cur) {
		if (typeid(*cur) == typeid(KiteDynamicObject)) {
			// Case 1: dynamic loading
			if (loader) {
				try {
					KiteReference<KiteObject> (*func)(VM &, KiteReference<KiteObject>) = 
						(KiteReference<KiteObject> (*)(VM &, KiteReference<KiteObject>))loader->findFunction(((*cur).type + string("_") + id).c_str());
					return KiteReference<KiteObject>(new KiteCode(func));
				} catch (KiteReference<KiteObject> &exc) {
					// Exception caught, ignore
				}
			} 

			// Case 2: normal user-written Kite code
			if (((KiteDynamicObject*)(cur))->syms.count(id) > 0) {
				return ((KiteDynamicObject*)(cur))->syms[id];
			}
		} else {
			if (cur->getSyms().count(id) > 0) {
				return cur->getSyms()[id];
			}
		}

		if (!staticobjs) {
			staticobjs = true;
			cur = &*(((KiteDynamicObject&)(*cur)).classinfo);
		} else {
			cur = (KiteObject*)cur->getInherits().getObj();
		}
	}

	// We can't find the object, so time to create it.
	return syms[id];
}
